import datetime
from dateutil.tz import tzutc
from unittest.mock import create_autospec

import pytest

import smarttub

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="mock_account")
def mock_account(mock_api):
    account = create_autospec(smarttub.Account, instance=True)
    return account


@pytest.fixture(name="spa")
def spa(mock_api, mock_account):
    spa = smarttub.Spa(mock_api, mock_account, id="id1", brand="brand1", model="model1")
    return spa


async def test_spa(spa):
    assert str(spa)


async def test_get_status(mock_api, spa):
    mock_api.request.return_value = {
        "ambientTemperature": 65.6,
        "blowoutCycle": "INACTIVE",
        "cleanupCycle": "INACTIVE",
        "current": {"average": 0.0, "kwh": 0.213, "max": 0.0, "min": 0.0, "value": 0.0},
        "date": "2021-02-21",
        "demoMode": "DISABLED",
        "dipSwitches": 8,
        "displayTemperatureFormat": "FAHRENHEIT",
        "error": {"code": 0, "description": None, "title": "All Clear"},
        "errorCode": 0,
        "fieldsLastUpdated": {
            "cfstEvent": None,
            "errEvent": "2021-02-17T09:10:31.059Z",
            "heatMode": "2020-07-09T19:40:01.883Z",
            "locEvent": "2021-02-21T01:20:41.956Z",
            "online": "2021-02-21T21:28:41.791Z",
            "rpstEvent": "2021-02-21T21:32:36.053Z",
            "setTemperature": "2021-02-20T03:10:00.525Z",
            "sp2stEvent": "2021-02-21T07:24:08.119Z",
            "spstEvent": "2021-02-21T20:40:09.330Z",
            "uv": "2021-02-21T18:48:34.699Z",
            "uvOnDemand": "2021-02-13T04:36:06.268Z",
            "wcstEvent": "2021-02-21T21:30:15.089Z",
        },
        "flowSwitch": "OPEN",
        "heatMode": "AUTO",
        "heater": "OFF",
        "highTemperatureLimit": 36.1,
        "lastUpdated": "2021-02-21T21:32:36.215Z",
        "lights": None,
        "locks": {
            "access": "UNLOCKED",
            "maintenance": "UNLOCKED",
            "spa": "UNLOCKED",
            "temperature": "UNLOCKED",
        },
        "online": True,
        "ozone": "OFF",
        "primaryFiltration": {
            "cycle": 1,
            "duration": 4,
            "lastUpdated": "2021-01-20T11:38:57.014Z",
            "mode": "NORMAL",
            "startHour": 2,
            "status": "INACTIVE",
        },
        "pumps": None,
        "secondaryFiltration": {
            "lastUpdated": "2020-07-09T19:39:52.961Z",
            "mode": "AWAY",
            "status": "INACTIVE",
        },
        "setTemperature": 38.3,
        "state": "NORMAL",
        "time": "14:45:00",
        "timeFormat": "HOURS_12",
        "timeSet": None,
        "timezone": None,
        "uv": "OFF",
        "uvOnDemand": "OFF",
        "versions": {"balboa": "1.06", "controller": "1.28", "jacuzziLink": "53"},
        "water": {
            "oxidationReductionPotential": 604,
            "ph": 7.01,
            "temperature": 38.3,
            "temperatureLastUpdated": "2021-02-21T16:40:10.054Z",
            "turbidity": 0.01,
        },
        "watercare": None,
    }

    status = await spa.get_status()
    assert status.state == "NORMAL"
    assert str(status)
    assert len(status.locks) == 4

    pf = status.primary_filtration
    assert pf.mode == status.primary_filtration.PrimaryFiltrationMode.NORMAL
    assert pf.duration == 4
    assert pf.start_hour == 2
    assert pf.status == status.CycleStatus.INACTIVE

    await pf.set(start_hour=5)
    mock_api.request.assert_called_with(
        "PATCH",
        f"spas/{spa.id}/config",
        body={
            "primaryFiltrationConfig": {
                "cycle": pf.cycle,
                "duration": pf.duration,
                "mode": pf.mode.name,
                "startHour": 5,
            }
        },
    )


async def test_get_status_full(mock_api, spa):
    mock_api.request.return_value = {
        "ambientTemperature": 65.6,
        "blowoutCycle": "INACTIVE",
        "cleanupCycle": "INACTIVE",
        "current": {"average": 9.5, "kwh": 0.375, "max": 9.6, "min": 9.4, "value": 9.5},
        "date": "2021-03-07",
        "demoMode": "DISABLED",
        "dipSwitches": 8,
        "displayTemperatureFormat": "FAHRENHEIT",
        "error": {"code": 0, "description": None, "title": "All Clear"},
        "errorCode": 0,
        "fieldsLastUpdated": {
            "cfstEvent": "2021-03-07T08:36:52.908Z",
            "errEvent": "2021-03-07T08:00:29.813Z",
            "heatMode": "2021-02-22T07:21:26.598Z",
            "locEvent": "2021-03-07T01:20:34.262Z",
            "online": "2021-03-07T22:04:13.477Z",
            "rpstEvent": "2021-03-07T22:05:21.288Z",
            "setTemperature": "2021-02-26T05:00:27.580Z",
            "sp2stEvent": "2021-03-07T08:36:54.867Z",
            "spstEvent": "2021-03-07T21:39:11.540Z",
            "uv": "2021-03-07T20:01:23.870Z",
            "uvOnDemand": "2021-02-13T04:36:06.268Z",
            "wcstEvent": "2021-03-07T21:59:14.378Z",
        },
        "flowSwitch": "OPEN",
        "heatMode": "AUTO",
        "heater": "OFF",
        "highTemperatureLimit": 38.9,
        "lastUpdated": "2021-03-07T22:05:21.440Z",
        "lights": [
            {
                "color": {"blue": 0, "green": 0, "red": 0, "white": 0},
                "cycleSpeed": 0,
                "intensity": 0,
                "mode": "OFF",
                "zone": 1,
            }
        ],
        "location": {"accuracy": 1053.0, "latitude": 27.129, "longitude": -27.906},
        "locks": {
            "access": "UNLOCKED",
            "maintenance": "UNLOCKED",
            "spa": "LOCKED",
            "temperature": "UNLOCKED",
        },
        "online": True,
        "ozone": "OFF",
        "primaryFiltration": {
            "cycle": 1,
            "duration": 4,
            "lastUpdated": "2021-02-24T02:55:47.180Z",
            "mode": "NORMAL",
            "startHour": 2,
            "status": "INACTIVE",
        },
        "pumps": [
            {
                "current": None,
                "id": "P2",
                "speed": "ONE_SPEED",
                "state": "OFF",
                "type": "JET",
            },
            {
                "current": None,
                "id": "P1",
                "speed": "ONE_SPEED",
                "state": "OFF",
                "type": "JET",
            },
            {
                "current": None,
                "id": "CP",
                "speed": "ONE_SPEED",
                "state": "OFF",
                "type": "CIRCULATION",
            },
        ],
        "secondaryFiltration": {
            "lastUpdated": "2021-03-04T16:47:29.882Z",
            "mode": "AWAY",
            "status": "INACTIVE",
        },
        "setTemperature": 38.3,
        "state": "NORMAL",
        "time": "14:05:00",
        "timeFormat": "HOURS_12",
        "timeSet": None,
        "timezone": None,
        "uv": "OFF",
        "uvOnDemand": "OFF",
        "versions": {"balboa": "1.06", "controller": "1.28", "jacuzziLink": "53"},
        "water": {
            "oxidationReductionPotential": 604,
            "ph": 7.01,
            "temperature": 38.9,
            "temperatureLastUpdated": "2021-03-07T22:04:15.686Z",
            "turbidity": 0.01,
        },
        "watercare": None,
    }

    status = await spa.get_status_full()
    assert len(status.pumps) == 3
    for pump in status.pumps:
        assert isinstance(pump, smarttub.SpaPump)
    assert len(status.lights) == 1
    for light in status.lights:
        assert isinstance(light, smarttub.SpaLight)

    assert status.spa == spa
    assert len(status.locks) == 4

    await status.locks["access"].lock()
    mock_api.request.assert_called_with(
        "POST", "spas/id1/lock", {"type": "ACCESS", "code": "0772"}
    )

    await status.locks["spa"].unlock()
    mock_api.request.assert_called_with(
        "POST", "spas/id1/unlock", {"type": "SPA", "code": "0772"}
    )


async def test_get_pumps(mock_api, spa):
    mock_api.request.return_value = {
        "pumps": [
            {
                "id": "pid1",
                "speed": "speed1",
                "state": "OFF",
                "type": "CIRCULATION",
            }
        ]
    }
    pumps = await spa.get_pumps()
    assert len(pumps) == 1
    pump = pumps[0]
    assert pump.id == "pid1"


async def test_get_lights(mock_api, spa):
    mock_api.request.return_value = {
        "lights": [
            {
                "color": {"blue": 0, "green": 0, "red": 0, "white": 0},
                "intensity": 0,
                "mode": "OFF",
                "zone": 1,
            }
        ]
    }
    lights = await spa.get_lights()
    assert len(lights) == 1
    light = lights[0]
    assert light.zone == 1


async def test_get_errors(mock_api, spa):
    mock_api.request.return_value = {
        "content": [
            {
                "code": 11,
                "title": "Flow Switch Stuck Open",
                "description": None,
                "createdAt": "2019-12-11T18:51:10.123Z",
                "updatedAt": "2020-07-14T19:00:50.705Z",
                "active": True,
                "errorType": "TUB_ERROR",
            }
        ],
        "pageable": {
            "sort": {"sorted": True, "unsorted": False, "empty": False},
            "pageSize": 10,
            "pageNumber": 0,
            "offset": 0,
            "paged": True,
            "unpaged": False,
        },
        "last": True,
        "totalPages": 1,
        "totalElements": 1,
        "first": True,
        "sort": {"sorted": True, "unsorted": False, "empty": False},
        "numberOfElements": 1,
        "size": 10,
        "number": 0,
        "empty": False,
    }
    errors = await spa.get_errors()
    assert len(errors) == 1
    error = errors[0]
    assert str(error)
    assert error.title == "Flow Switch Stuck Open"


async def test_get_reminders(mock_api, spa):
    mock_api.request.return_value = {
        "reminders": [
            {
                "id": "id1",
                "lastUpdated": "2020-07-09T06:42:53.857Z",
                "name": "name1",
                "remainingDuration": 23,
                "snoozed": False,
                "state": "INACTIVE",
            }
        ]
    }
    reminders = await spa.get_reminders()
    assert len(reminders) == 1
    reminder = reminders[0]
    assert reminder.id == "id1"
    assert reminder.name == "name1"


async def test_get_debug_status(mock_api, spa):
    mock_api.request.return_value = {
        "debugStatus": {
            "battery": {"percentCharge": 82.2, "voltage": 4.04},
            "freeMemory": 45520,
            "lastResetReason": "RESET_REASON_USER",
            "powerStatus": "DC",
            "resetCount": 133,
            "signal": {"quality": 37, "strength": -89},
            "uptime": {
                "connection": 2059053,
                "system": 2059161,
                "tubController": 2059159,
            },
        }
    }

    debug_status = await spa.get_debug_status()
    assert debug_status is not None


async def test_get_energy_usage(mock_api, spa):
    mock_api.request.return_value = {"buckets": []}
    usage = await spa.get_energy_usage(
        smarttub.Spa.EnergyUsageInterval.DAY,
        datetime.date(2021, 1, 1),
        datetime.date(2021, 1, 31),
    )
    mock_api.request.assert_called_with(
        "POST",
        f"spas/{spa.id}/energyUsage",
        {"start": "2021-01-01", "end": "2021-01-31", "interval": "DAY"},
    )
    assert usage == []


async def test_set_heat_mode(mock_api, spa):
    await spa.set_heat_mode(smarttub.Spa.HeatMode.AUTO)
    mock_api.request.assert_called_with(
        "PATCH", f"spas/{spa.id}/config", {"heatMode": "AUTO"}
    )


async def test_set_temperature(mock_api, spa):
    await spa.set_temperature(38.3)
    mock_api.request.assert_called_with(
        "PATCH", f"spas/{spa.id}/config", {"setTemperature": 38.3}
    )


async def test_toggle_clearray(mock_api, spa):
    await spa.toggle_clearray()
    mock_api.request.assert_called_with("POST", f"spas/{spa.id}/clearray/toggle", None)


async def test_set_temperature_format(mock_api, spa):
    await spa.set_temperature_format(smarttub.Spa.TemperatureFormat.FAHRENHEIT)
    mock_api.request.assert_called_with(
        "POST", f"spas/{spa.id}/config", {"displayTemperatureFormat": "FAHRENHEIT"}
    )


async def test_set_date_time(mock_api, spa):
    with pytest.raises(ValueError):
        await spa.set_date_time()

    await spa.set_date_time(date=datetime.date(2021, 1, 1))
    mock_api.request.assert_called_with(
        "POST", f"spas/{spa.id}/config", {"dateTimeConfig": {"date": "2021-01-01"}}
    )
    await spa.set_date_time(time=datetime.time(12, 45))
    mock_api.request.assert_called_with(
        "POST", f"spas/{spa.id}/config", {"dateTimeConfig": {"time": "12:45"}}
    )


async def test_secondary_filtration_cycle(mock_api, spa):
    cycle = smarttub.SpaSecondaryFiltrationCycle(
        spa,
        **{
            "lastUpdated": "2020-07-09T19:39:52.961Z",
            "mode": "AWAY",
            "status": "INACTIVE",
        },
    )
    assert cycle.last_updated == datetime.datetime(
        2020, 7, 9, 19, 39, 52, 961000, tzinfo=tzutc()
    )
    assert (
        cycle.mode == smarttub.SpaSecondaryFiltrationCycle.SecondaryFiltrationMode.AWAY
    )
    assert cycle.status == smarttub.SpaSecondaryFiltrationCycle.CycleStatus.INACTIVE

    await cycle.set_mode(
        smarttub.SpaSecondaryFiltrationCycle.SecondaryFiltrationMode.FREQUENT
    )
    mock_api.request.assert_called_with(
        "PATCH", f"spas/{spa.id}/config", {"secondaryFiltrationConfig": "FREQUENT"}
    )



import asyncio
import datetime
from enum import Enum
import logging
import time
from typing import List

import aiohttp
import dateutil.parser
from inflection import underscore
import jwt

logger = logging.getLogger(__name__)


class SmartTub:
    """Interface to the SmartTub API"""

    AUTH_AUDIENCE = "https://api.operation-link.com/"
    AUTH_URL = "https://smarttub.auth0.com/oauth/token"
    AUTH_CLIENT_ID = "dB7Rcp3rfKKh0vHw2uqkwOZmRb5WNjQC"
    AUTH_REALM = "Username-Password-Authentication"
    AUTH_ACCOUNT_ID_KEY = "http://operation-link.com/account_id"
    AUTH_GRANT_TYPE = "http://auth0.com/oauth/grant-type/password-realm"
    AUTH_SCOPE = "openid email offline_access User Admin"

    API_BASE = "https://api.smarttub.io"

    def __init__(self, session: aiohttp.ClientSession = None):
        self.logged_in = False
        self._session = session or aiohttp.ClientSession()

    async def login(self, username: str, password: str):
        """Authenticate to SmartTub

        This method must be called before any useful work can be done.

        username -- the email address for the SmartTub account
        password -- the password for the SmartTub account
        """

        # https://auth0.com/docs/api-auth/tutorials/password-grant
        r = await self._session.post(
            self.AUTH_URL,
            json={
                "audience": self.AUTH_AUDIENCE,
                "client_id": self.AUTH_CLIENT_ID,
                "grant_type": self.AUTH_GRANT_TYPE,
                "realm": self.AUTH_REALM,
                "scope": self.AUTH_SCOPE,
                "username": username,
                "password": password,
            },
        )
        if r.status == 403:
            raise LoginFailed(r.text)

        r.raise_for_status()
        j = await r.json()

        self._set_access_token(j["access_token"])
        self.refresh_token = j["refresh_token"]
        assert j["token_type"] == "Bearer"

        self.account_id = self.access_token_data[self.AUTH_ACCOUNT_ID_KEY]
        self.logged_in = True

        logger.debug(f"login successful, username={username}")

    @property
    def _headers(self):
        return {"Authorization": f"Bearer {self.access_token}"}

    async def _require_login(self):
        if not self.logged_in:
            raise RuntimeError("not logged in")
        if self.token_expires_at <= time.time():
            await self._refresh_token()

    def _set_access_token(self, token):
        self.access_token = token
        self.access_token_data = jwt.decode(
            self.access_token,
            algorithms=["HS256"],
            options={"verify_signature": False, "verify": False},
        )
        self.token_expires_at = self.access_token_data["exp"]

    async def _refresh_token(self):
        # https://auth0.com/docs/tokens/guides/use-refresh-tokens
        r = await self._session.post(
            self.AUTH_URL,
            json={
                "grant_type": "refresh_token",
                "client_id": self.AUTH_CLIENT_ID,
                "refresh_token": self.refresh_token,
            },
        )
        r.raise_for_status()
        j = await r.json()
        self._set_access_token(j["access_token"])
        logger.debug("token refresh successful")

    async def request(self, method, path, body=None):
        """Generic method for making an authenticated request to the API

        This is used by resource objects associated with this API object
        """

        await self._require_login()

        r = await self._session.request(
            method, f"{self.API_BASE}/{path}", headers=self._headers, json=body
        )

        try:
            r.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise APIError(e)

        if int(r.headers["content-length"]) == 0:
            ret = None
        else:
            ret = await r.json()

        logger.debug(f"{method} {path} successful: {ret}")

        return ret

    async def get_account(self) -> "Account":
        """Retrieve the SmartTub account of the authenticated user"""

        j = await self.request("GET", f"accounts/{self.account_id}")
        account = Account(self, **j)
        logger.debug(f"get_account successful: {j}")

        return account


class Account:
    def __init__(self, api: SmartTub, **properties):
        self._api = api
        self.id = properties["id"]
        self.email = properties["email"]
        self.properties = properties

    async def get_spas(self):
        return await asyncio.gather(
            *[
                self.get_spa(spa["id"])
                for spa in (await self._api.request("GET", f"spas?ownerId={self.id}"))[
                    "content"
                ]
            ]
        )

    async def get_spa(self, spa_id: str):
        return Spa(self._api, self, **await self._api.request("GET", f"spas/{spa_id}"))

    def __str__(self):
        return f"<Account {self.email}>"


class Spa:
    HeatMode = Enum("HeatMode", "ECONOMY DAY AUTO")
    TemperatureFormat = Enum("TemperatureFormat", "FAHRENHEIT CELSIUS")
    EnergyUsageInterval = Enum("EnergyUsageInterval", "DAY MONTH")

    def __init__(self, api: SmartTub, account: Account, **properties):
        self._api = api
        self.account = account
        self.id = properties["id"]
        self.brand = properties["brand"]
        self.model = properties["model"]
        self.properties = properties

        self.name = f"{self.brand} {self.model}"

    async def request(self, method, resource: str, body=None):
        return await self._api.request(method, f"spas/{self.id}/{resource}", body)

    async def get_status(self) -> "SpaState":
        """Query the status of the spa."""
        return SpaState(self, **await self.request("GET", "status"))

    async def get_pumps(self) -> List["SpaPump"]:
        return [
            SpaPump(self, **pump_info)
            for pump_info in (await self.request("GET", "pumps"))["pumps"]
        ]

    async def get_lights(self) -> List["SpaLight"]:
        return [
            SpaLight(self, **light_info)
            for light_info in (await self.request("GET", "lights"))["lights"]
        ]

    async def get_errors(self) -> List["SpaError"]:
        return [
            SpaError(self, **error_info)
            for error_info in (await self.request("GET", "errors"))["content"]
        ]

    async def get_reminders(self) -> List["SpaReminder"]:
        # API returns both 'reminders' and 'filters', both seem to be identical
        return [
            SpaReminder(self, **reminder_info)
            for reminder_info in (await self.request("GET", "reminders"))["reminders"]
        ]

    async def get_status_full(self) -> "SpaStateFull":
        """Retrieves the state of lights and pumps in addition to what get_status does."""
        full_status = await self.request("GET", "fullStatus")
        return SpaStateFull(self, full_status)

    async def get_debug_status(self) -> dict:
        return (await self.request("GET", "debugStatus"))["debugStatus"]

    async def get_energy_usage(
        self,
        interval: EnergyUsageInterval,
        start_date: datetime.date,
        end_date: datetime.date,
    ) -> list:
        body = {
            "start": start_date.isoformat(),
            "end": end_date.isoformat(),
            "interval": interval.name,
        }
        return (await self.request("POST", "energyUsage", body))["buckets"]

    async def set_heat_mode(self, mode: HeatMode):
        body = {"heatMode": mode.name}
        await self.request("PATCH", "config", body)

    async def set_temperature(self, temp_c: float):
        body = {
            # responds with 500 if given more than 1 decimal point
            "setTemperature": round(temp_c, 1)
        }
        await self.request("PATCH", "config", body)

    async def toggle_clearray(self):
        await self.request("POST", "clearray/toggle")

    async def set_temperature_format(self, temperature_format: TemperatureFormat):
        body = {"displayTemperatureFormat": temperature_format.name}
        await self.request("POST", "config", body)

    async def set_date_time(
        self, date: datetime.date = None, time: datetime.time = None
    ):
        """Set the spa date, time, or both"""

        if date is None and time is None:
            raise ValueError("at least one of date or time must be specified")
        config = {}
        if date is not None:
            config["date"] = date.isoformat()
        if time is not None:
            config["time"] = time.isoformat("minutes")
        body = {"dateTimeConfig": config}
        await self.request("POST", "config", body)

    def __str__(self):
        return f"<Spa {self.id}>"


class SpaState:
    CycleStatus = Enum("CycleStatus", "INACTIVE ACTIVE")

    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.properties = properties.copy()
        self._prop("ambientTemperature")
        self._prop("blowoutCycle", constructor=lambda x: self.CycleStatus[x])
        self._prop("cleanupCycle", constructor=lambda x: self.CycleStatus[x])
        self._prop("current")
        self._prop("date", constructor=dateutil.parser.isoparse)
        self._prop("demoMode")
        self._prop("dipSwitches")
        self._prop("displayTemperatureFormat")
        self._prop("error")
        self._prop("errorCode")
        self._prop(
            "fieldsLastUpdated",
            constructor=lambda d: {
                k: dateutil.parser.isoparse(v) if v is not None else None
                for k, v in d.items()
            },
        )
        self._prop("flowSwitch")
        self._prop("heatMode", constructor=lambda x: Spa.HeatMode[x])
        self._prop("heater")
        self._prop("highTemperatureLimit")
        self._prop("lastUpdated", constructor=dateutil.parser.isoparse)
        self._prop("lights")  # seems to be None even when there are lights?
        self._prop("location")
        self._prop(
            "locks",
            constructor=lambda x: {
                k: SpaLock(self.spa, kind=k, state=v) for k, v in x.items()
            },
        )
        self._prop("online")
        self._prop("ozone")
        self._prop(
            "primaryFiltration",
            constructor=lambda p: SpaPrimaryFiltrationCycle(self.spa, **p),
        )
        self._prop(
            "secondaryFiltration",
            constructor=lambda p: SpaSecondaryFiltrationCycle(self.spa, **p),
        )
        self._prop("setTemperature")
        self._prop("state")
        self._prop("time", constructor=datetime.time.fromisoformat)
        self._prop("timeFormat")
        self._prop("timeSet")  # ?
        self._prop("timezone")  # ?
        self._prop("uv")
        self._prop("uvOnDemand")
        self._prop("versions")
        self._prop("water", constructor=lambda p: SpaWaterState(self.spa, **p))
        self._prop("watercare")

    def _prop(
        self, json_key, instance_variable_name=None, optional=True, constructor=None
    ):
        if instance_variable_name is None:
            instance_variable_name = underscore(json_key)
        if json_key in self.properties:
            if constructor is None:
                setattr(self, instance_variable_name, self.properties[json_key])
            else:
                setattr(
                    self, instance_variable_name, constructor(self.properties[json_key])
                )
        else:
            setattr(self, instance_variable_name, None)

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.properties}>"


class SpaStateFull(SpaState):
    def __init__(self, spa: Spa, state: dict):
        super().__init__(spa, **state)
        self.lights = [
            SpaLight(spa, **light_props) for light_props in self.properties["lights"]
        ]
        self.pumps = [
            SpaPump(spa, **pump_props) for pump_props in self.properties["pumps"]
        ]


class SpaWaterState(SpaState):
    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.properties = properties.copy()

        self._prop("temperature")
        self._prop("temperatureLastUpdated", constructor=dateutil.parser.isoparse)


class SpaPrimaryFiltrationCycle(SpaState):
    PrimaryFiltrationMode = Enum("PrimaryFiltrationMode", "NORMAL")

    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.properties = properties.copy()

        self._prop("cycle")
        self._prop("duration")
        self._prop("lastUpdated", constructor=dateutil.parser.isoparse)
        self._prop("mode", constructor=lambda x: self.PrimaryFiltrationMode[x])
        self._prop("startHour")
        self._prop("status", constructor=lambda x: self.CycleStatus[x])

    async def set(self, cycle=None, duration=None, mode=None, start_hour=None):
        body = {
            "primaryFiltrationConfig": {
                "cycle": cycle if cycle is not None else self.cycle,
                "duration": duration if duration is not None else self.duration,
                "mode": mode.name if mode is not None else self.mode.name,
                "startHour": start_hour if start_hour is not None else self.start_hour,
            }
        }
        await self.spa.request("PATCH", "config", body)


class SpaSecondaryFiltrationCycle(SpaState):
    SecondaryFiltrationMode = Enum(
        "SecondaryFiltrationMode", "AWAY FREQUENT INFREQUENT"
    )

    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.properties = properties.copy()

        self._prop("lastUpdated", constructor=dateutil.parser.isoparse)
        self._prop("mode", constructor=lambda x: self.SecondaryFiltrationMode[x])
        self._prop("status", constructor=lambda x: self.CycleStatus[x])

    async def set_mode(self, mode: SecondaryFiltrationMode):
        body = {"secondaryFiltrationConfig": mode.name}
        await self.spa.request("PATCH", "config", body)


class SpaPump:
    PumpState = Enum("PumpState", "OFF LOW HIGH")
    PumpType = Enum("PumpType", "BLOWER CIRCULATION JET")

    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.id = properties["id"]
        self.speed = properties["speed"]
        self.state = self.PumpState[properties["state"]]
        self.type = self.PumpType[properties["type"]]
        self.properties = properties

    async def toggle(self):
        await self.spa.request("POST", f"pumps/{self.id}/toggle")

    def __str__(self):
        return f"<SpaPump {self.id}: {self.type.name}={self.state.name}>"


class SpaLight:
    LightMode = Enum(
        "LightMode",
        "PURPLE ORANGE RED YELLOW GREEN AQUA BLUE WHITE HIGH_SPEED_COLOR_WHEEL FULL_DYNAMIC_RGB OFF",
    )

    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.zone = properties["zone"]

        color = properties["color"]
        self.red = color["red"]
        self.green = color["green"]
        self.blue = color["blue"]
        self.white = color["white"]

        self.intensity = properties["intensity"]
        self.mode = self.LightMode[properties["mode"]]
        self.properties = properties

    async def set_mode(self, mode: LightMode, intensity: int):
        assert (intensity == 0) == (mode == self.LightMode.OFF)

        body = {
            "intensity": intensity,
            "mode": mode.name,
        }
        await self.spa.request("PATCH", f"lights/{self.zone}", body)

    async def turn_off(self):
        await self.set_mode(self.LightMode.OFF, 0)

    def __str__(self):
        return f"<SpaLight {self.zone}: {self.mode.name} (R {self.red}/G {self.green}/B {self.blue}/W {self.white}) @ {self.intensity}>"


class SpaReminder:
    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.id = properties["id"]
        self.name = properties["name"]
        self.remaining_days = properties["remainingDuration"]
        self.snoozed = properties["snoozed"]
        self.state = properties["state"]
        self.last_updated = None

        last_updated_str = properties.get("lastUpdated")
        if last_updated_str is not None:
            self.last_updated = dateutil.parser.isoparse(last_updated_str)

    async def snooze(self, days: int):
        body = {"remainingDuration": days}
        await self.spa.request("PATCH", f"reminders/{self.id}", body)

    async def reset(self, days: int):
        body = {"remainingDuration": days, "reset": True}
        await self.spa.request("PATCH", f"reminders/{self.id}", body)

    def __str__(self):
        return f"<SpaReminder {self.id}: {self.state}/{self.remaining_days}/{self.snoozed}>"


class SpaError:
    def __init__(self, spa: Spa, **properties):
        self.spa = spa
        self.code = properties["code"]
        self.title = properties["title"]
        self.description = properties["description"]
        self.created_at = dateutil.parser.isoparse(properties["createdAt"])
        self.updated_at = dateutil.parser.isoparse(properties["updatedAt"])
        self.active = properties["active"]
        self.error_type = properties["errorType"]

    def __str__(self):
        return f"<SpaError {self.title}>"


class SpaLock:
    CODE = "0772"

    def __init__(self, spa: Spa, kind: str, state: str):
        self.spa = spa
        self.kind = kind
        self.state = state

    async def lock(self):
        if self.state != "LOCKED":
            await self.spa.request(
                "POST",
                "lock",
                {
                    "type": self.kind.upper(),
                    "code": self.CODE,
                },
            )

    async def unlock(self):
        if self.state != "UNLOCKED":
            await self.spa.request(
                "POST",
                "unlock",
                {
                    "type": self.kind.upper(),
                    "code": self.CODE,
                },
            )

    def __str__(self):
        return f"<SpaLock {self.kind}: {self.state}>"


class LoginFailed(RuntimeError):
    pass


class APIError(RuntimeError):
    pass



import argparse
import asyncio
import datetime
import logging
from pprint import pprint
import sys

import aiohttp

from . import SmartTub, SpaLight


async def info_command(spas, args):
    for spa in spas:
        print(f"= Spa '{spa.name}' =\n")
        if args.all or args.status or args.location or args.locks:
            status = await spa.get_status()

        if args.all or args.status:
            print("== Status ==")
            status_dict = status.properties.copy()
            # redact location for privacy
            location = status_dict.pop("location")
            pprint(status_dict)
            print()

        if args.location:
            # not included in --all
            print(
                f"Location: {location['latitude']} {location['longitude']} (accuracy: {location['accuracy']})\n"
            )

        if args.all or args.pumps:
            print("== Pumps ==")
            for pump in await spa.get_pumps():
                print(pump)
            print()

        if args.all or args.lights:
            print("== Lights ==")
            for light in await spa.get_lights():
                print(light)
            print()

        if args.all or args.errors:
            print("== Errors ==")
            for error in await spa.get_errors():
                print(error)
            print()

        if args.all or args.reminders:
            print("== Reminders ==")
            for reminder in await spa.get_reminders():
                print(reminder)
            print()

        if args.all or args.locks:
            print("== Locks ==")
            for lock in status.locks.values():
                print(lock)
            print()

        if args.all or args.energy:
            print("== Energy usage ==")
            energy_usage_day = spa.get_energy_usage(
                spa.EnergyUsageInterval.DAY,
                end_date=datetime.date.today(),
                start_date=datetime.date.today() - datetime.timedelta(days=7),
            )
            pprint(await energy_usage_day)
            print()

        if args.all or args.debug:
            debug_status = await spa.get_debug_status()
            print("== Debug status ==")
            pprint(debug_status)
            print()


async def set_command(spas, args):
    for spa in spas:
        if args.temperature:
            await spa.set_temperature(args.temperature)

        if args.light_mode:
            for light in await spa.get_lights():
                if args.verbosity > 0:
                    print(light)
                mode = light.LightMode[args.light_mode]
                if mode == light.LightMode.OFF:
                    await light.set_mode(mode, 0)
                else:
                    await light.set_mode(mode, 50)

        if args.snooze_reminder:
            reminder_id, days = args.snooze_reminder
            days = int(days)
            reminder = next(
                reminder
                for reminder in await spa.get_reminders()
                if reminder.id == reminder_id
            )
            await reminder.snooze(days)

        if args.reset_reminder:
            reminder_id, days = args.reset_reminder
            days = int(days)
            reminder = next(
                reminder
                for reminder in await spa.get_reminders()
                if reminder.id == reminder_id
            )
            await reminder.reset(days)

        if args.lock:
            status = await spa.get_status()
            lock = status.locks[args.lock.lower()]
            await lock.lock()
            print("OK")

        if args.unlock:
            status = await spa.get_status()
            lock = status.locks[args.unlock.lower()]
            await lock.unlock()
            print("OK")


async def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-u", "--username", required=True, help="SmartTub account email"
    )
    parser.add_argument(
        "-p", "--password", required=True, help="SmartTub account password"
    )
    parser.add_argument("-v", "--verbosity", action="count", default=0)
    subparsers = parser.add_subparsers()

    info_parser = subparsers.add_parser("info", help="Show information about the spa")
    info_parser.set_defaults(func=info_command)
    info_parser.add_argument(
        "-a", "--all", action="store_true", help="Show all info except location"
    )
    info_parser.add_argument("--spas", action="store_true")
    info_parser.add_argument("--status", action="store_true")
    info_parser.add_argument(
        "--location", action="store_true", help="Show GPS location"
    )
    info_parser.add_argument("--pumps", action="store_true")
    info_parser.add_argument("--lights", action="store_true")
    info_parser.add_argument("--errors", action="store_true")
    info_parser.add_argument("--reminders", action="store_true")
    info_parser.add_argument("--locks", action="store_true")
    info_parser.add_argument("--debug", action="store_true")
    info_parser.add_argument("--energy", action="store_true")

    set_parser = subparsers.add_parser("set", help="Change settings on the spa")
    set_parser.set_defaults(func=set_command)
    set_parser.add_argument(
        "-l", "--light_mode", choices=[mode.name for mode in SpaLight.LightMode]
    )
    set_parser.add_argument("-t", "--temperature", type=float)
    # TODO: should enforce types of str, int
    set_parser.add_argument(
        "--snooze-reminder",
        nargs=2,
        help="Snooze a reminder",
        metavar=("REMINDER_ID", "DAYS"),
    )
    # TODO: should enforce types of str, int
    set_parser.add_argument(
        "--reset-reminder",
        nargs=2,
        help="Reset a reminder",
        metavar=("REMINDER_ID", "DAYS"),
    )
    set_parser.add_argument("--lock", type=str)
    set_parser.add_argument("--unlock", type=str)

    args = parser.parse_args(argv)

    if args.verbosity > 1:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logging.basicConfig(level=log_level)

    async with aiohttp.ClientSession() as session:
        st = SmartTub(session)
        await st.login(args.username, args.password)

        account = await st.get_account()

        spas = await account.get_spas()
        await args.func(spas, args)


asyncio.run(main(sys.argv[1:]))



from unittest.mock import create_autospec

import pytest

import smarttub


ACCOUNT_ID = "account_id1"


@pytest.fixture
def mock_api():
    api = create_autospec(smarttub.SmartTub, instance=True)
    return api


@pytest.fixture
def mock_spa():
    spa = create_autospec(smarttub.Spa, instance=True)
    return spa



import pytest

import smarttub

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="account")
def account(mock_api):
    return smarttub.Account(mock_api, id="id1", email="email1")


async def test_account(account):
    assert str(account)


async def test_get_spas(mock_api, account):
    mock_api.request.side_effect = [
        {"content": [{"id": "sid1"}]},
        {"id": "sid1", "brand": "brand1", "model": "model1"},
    ]
    spas = await account.get_spas()
    assert len(spas) == 1
    spa = spas[0]
    assert spa.id == "sid1"



import pytest

from smarttub import SpaLight

pytestmark = pytest.mark.asyncio


@pytest.fixture
def lights(mock_spa):
    lights = [
        SpaLight(
            mock_spa,
            **{
                "color": {"blue": 0, "green": 0, "red": 0, "white": 0},
                "intensity": 0 if mode == SpaLight.LightMode.OFF else 50,
                "mode": mode.name,
                "zone": i + 1,
            },
        )
        for i, mode in enumerate(SpaLight.LightMode)
    ]
    return lights


async def test_light(mock_spa, lights):
    purple = lights[0]
    assert str(purple)
    assert purple.mode == SpaLight.LightMode.PURPLE
    await purple.set_mode(SpaLight.LightMode.RED, 50)
    mock_spa.request.assert_called_with(
        "PATCH", f"lights/{purple.zone}", {"intensity": 50, "mode": "RED"}
    )
    await purple.turn_off()
    mock_spa.request.assert_called_with(
        "PATCH", f"lights/{purple.zone}", {"intensity": 0, "mode": "OFF"}
    )



import aiohttp
import time

import jwt
import pytest

import smarttub

ACCOUNT_ID = "account_id1"

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="unauthenticated_api")
async def unauthenticated_api(aresponses):
    async with aiohttp.ClientSession() as session:
        yield smarttub.SmartTub(session)


@pytest.fixture(name="api")
async def api(unauthenticated_api, aresponses):
    api = unauthenticated_api
    aresponses.add(
        response={
            "access_token": jwt.encode(
                {api.AUTH_ACCOUNT_ID_KEY: ACCOUNT_ID, "exp": time.time() + 3600},
                "secret",
            ),
            "token_type": "Bearer",
            "refresh_token": "refresh1",
        }
    )
    await api.login("username1", "password1")
    return api


async def test_login(api, aresponses):
    assert api.account_id == ACCOUNT_ID
    assert api.logged_in is True


async def test_login_failed(api, aresponses):
    aresponses.add(response=aresponses.Response(status=403))
    with pytest.raises(smarttub.LoginFailed):
        await api.login("username", "password")


async def test_refresh_token(api, aresponses):
    now = time.time()
    api.token_expires_at = now
    aresponses.add(
        response={
            "access_token": jwt.encode(
                {api.AUTH_ACCOUNT_ID_KEY: ACCOUNT_ID, "exp": now + 3601},
                "secret",
            ),
        }
    )
    aresponses.add(response={"status": "OK"})
    response = await api.request("GET", "/")
    assert api.token_expires_at > now
    assert response.get("status") == "OK"


async def test_get_account(api, aresponses):
    aresponses.add(
        response={
            "id": "id1",
            "email": "email1",
        }
    )

    account = await api.get_account()
    assert account.id == "id1"
    assert account.email == "email1"


async def test_api_error(api, aresponses):
    aresponses.add(response=aresponses.Response(status=500))
    with pytest.raises(smarttub.APIError):
        await api.get_account()


async def test_not_logged_in(unauthenticated_api, aresponses):
    with pytest.raises(RuntimeError):
        await unauthenticated_api.request("GET", "/")


async def test_request(api, aresponses):
    aresponses.add(response=aresponses.Response(text=None, status=200))
    response = await api.request("GET", "/")
    assert response is None



import pytest

from smarttub import SpaReminder

pytestmark = pytest.mark.asyncio


@pytest.fixture
def reminders(mock_spa):
    reminders = [
        SpaReminder(
            mock_spa,
            **{
                "id": "WATER",
                "lastUpdated": "2021-03-04T08:00:45.330Z",
                "name": "Refresh Water",
                "remainingDuration": 0,
                "snoozed": False,
                "state": "INACTIVE",
            }
        ),
        SpaReminder(
            mock_spa,
            **{
                "id": "NO_LAST_UPDATED",
                # lastUpdated may be null, see
                # https://github.com/mdz/python-smarttub/issues/22
                "lastUpdated": None,
                "name": "Refresh Water",
                "remainingDuration": 0,
                "snoozed": False,
                "state": "INACTIVE",
            }
        ),
    ]
    return reminders


async def test_reminders(mock_spa, reminders):
    reminder = reminders[0]
    assert str(reminder)
    assert reminder.id == "WATER"
    await reminder.snooze(5)
    mock_spa.request.assert_called_with(
        "PATCH", "reminders/WATER", {"remainingDuration": 5}
    )
    mock_spa.reset_mock()
    await reminder.reset(365)
    mock_spa.request.assert_called_with(
        "PATCH", "reminders/WATER", {"remainingDuration": 365, "reset": True}
    )



import pytest

from smarttub import SpaPump

pytestmark = pytest.mark.asyncio


@pytest.fixture
def pumps(mock_spa):
    pumps = [
        SpaPump(
            mock_spa,
            **{
                "id": "pid1",
                "speed": "speed1",
                "state": "OFF"
                if pump_type == SpaPump.PumpType.CIRCULATION
                else SpaPump.PumpState.HIGH.name,
                "type": pump_type.name,
            },
        )
        for i, pump_type in enumerate(SpaPump.PumpType)
    ]
    return pumps


async def test_pump(mock_spa, pumps):
    circ = pumps[1]
    assert str(circ)
    assert circ.speed == "speed1"
    assert circ.state == SpaPump.PumpState.OFF
    assert circ.type == SpaPump.PumpType.CIRCULATION
    await circ.toggle()
    mock_spa.request.assert_called_with("POST", f"pumps/{circ.id}/toggle")



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="python-smarttub",
    version="0.0.29",
    author="Matt Zimmerman",
    author_email="mdz@alcor.net",
    description="API to query and control hot tubs using the SmartTub system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mdz/python-smarttub",
    packages=["smarttub"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "aiohttp>=3.7.4,<4",
        "inflection~=0.5.1",
        "pyjwt~=2.1.0",
        "python-dateutil~=2.8.1",
    ],
    # Note: tests require python >=3.8
    tests_require=[
        "pytest",
        "pytest-asyncio",
        "aresponses",
    ],
    python_requires=">=3.7",
)
